; Keyboard driver module
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Progress:
;	5/9/99	Rewrote.  The new keyboard driver is similar to Linux's, and
;		it can use Linux keymaps (after conversion by keyconv).
;
; Bugs:
;
;
; Todo:
;	Routine(s) to remap a key
;
;	Handle compose
;
;	Handle keys which are linked to a function
;
;	Caps/Num/Scroll lock LED's :)
;
;	CTRL-ALT-DEL for reboot.
;	CTRL-Z or CTRL-ESC for trap to debugger.
;	PAUSE to pause current task.
;	CTRL-PAUSE (BREAK) to kill current task.
;	(How about a routine to install traps/hotkeys?)
;
;	Virtual console support: Add keystrokes to whatever buffer [keybuf]
;	points to.  Reading them is more complicated, since the task reading
;	them might not be running in the current VC at the moment.  So we
;	have to figure out which VC the task is running in by examining the
;	event table (area right before TSS).  Actually this involves serial
;	lines, networks, etc... see Notes.txt.
;
;	Spice up readline: arrow keys, home, end, insert, delete, command
;	history (Also, move to tty module)
;

%define ncols	(keymap)
%define offset	(keymap+4)
%define table	(keymap+8)

; Action types
%define A_FUNCTION	0
%define A_ASCII		1
%define A_STRING	2
%define A_MODIFIER	3


	SECTION .text

init_kbd:
;
; Install IRQ1 handler and initialize typeahead buffer
;
	mov	ecx, 128/4		;Clear the key_table
	mov	edi, key_table
	xor	eax, eax
	cld
	rep stosd

	;; Locate address of 'kt_cols' array
	mov	eax, [ncols]
	mov	ebx, [offset]
	shl	eax, 10		; eax = eax * 128 * 8 = eax*2^(7+3)
				; eax = Size of table
	add	eax, table	; eax = address of string table
	add	eax, ebx	; eax = address of kt_cols[]
	mov	[kt_cols], eax

; Todo: For keys that are mapped to function calls, link them...

	callproc patch_int,	0x21, irq1_handler
	in	al, 0x21
	and	al, 11111101b
	out	0x21, al

	ret



irq1_handler:
;
; Every time a key is pressed or released, do scancode conversion.
;
	pusha
	xor	eax, eax
	in	al, 0x60

; %define PAUSE		;Uncomment to handle PAUSE properly
%ifdef PAUSE
	cmp	[pause_key_loop], byte 0
	jz	.not_pause_loop
	dec	byte [pause_key_loop]
	jmp	.end
.not_pause_loop
%endif
	cmp	al, 0xE0	;224 = extended
	jne	.not_extended
	mov	[extended_flag], byte 1			;;;;
	jmp  .end
.not_extended
	cmp	al, 0xE1	;E1 = pause
	jne	.not_pause
.pause
%ifndef PAUSE
	call reboot	; For now, pause will be the REBOOT key :)
%else
	not byte [key_table+KEY_PAUSE]	;Toggle PAUSE key
	mov byte [pause_key_loop], 5	;Ignore next 5 scancodes
	mov	al, 119			;"fake" scancode for PAUSE
	cmp byte [key_table+KEY_PAUSE], 0
	je	.1			;Press
	and	al, 0x80		;Release
.1
%endif
.not_pause

.decode
	mov	ebx, eax
	and	bl, 0x7F	;BL has bit 7 off
;; Convert extended scancodes to "fake" scancodes
	cmp byte [extended_flag], 0
	jz	.not_ex
	mov	edi, xlat.from			; translation table
	mov	ecx, xlat.to - xlat.from	; size of table
	xchg	al, bl
	cld
	repne	scasb
	jne	.err		;Unexpected key?
	neg	ecx
	add	ecx, xlat.to - xlat.from - 1
	mov	al, [xlat.to + ecx]		;Translate
.err	xchg	al, bl
	mov byte [extended_flag], 0

.not_ex	xor	ah, ah
	test	al, 0x80
	jnz	.rel		;Release
.press	inc	ah
.rel	call	process_keystroke
; if toggle (caps/num/scroll), toggle a flag

.finish	mov	[key_table+ebx], ah	;Update the key's status

.end	irqack	1
	sti
	popa
	iretd



process_keystroke:
;
; Decode a character and add it to the typeahead buffer
;
; Parameters:
;	AH	1=press, 0=release
;	EBX (really just BL) = scancode (from irq1_handler)
;
	pusha
	lea	esi, [table + ebx*8]	; ESI = table entry, in 1st column
	;; if modifier, update the modifer bitmask
	cmp byte [esi], A_MODIFIER
	jnz	.1
	mov	bh, [esi+4]	;BH = bitmask for the modifier
	or	ah, ah
	jz	.rel
	or	[mods], bh	;key pressed - turn bit on
	jmp	.end
.rel	not	bh		;key released - turn bit off
	and	[mods], bh
	jmp	.end
.1

	;; Is the key being released?  Ignore it.
	or	ah, ah
	jz near .end


;; Select column, based on modifier bitmask
	mov	al, [mods]
	mov	ecx, [ncols]
	mov	edi, [kt_cols]
	cld
	repne	scasb
	jne near .end			; Unused combination? Ignore the key.
	not	ecx
	add	ecx, [ncols]
	shl	ecx, 10			; ecx = ecx * (128*8 bytes/keymap)
	add	esi, ecx		; esi = table entry, in the right column
;callproc print_hex, [esi]
;callproc print_hex, [esi+4]
;cr

	;; Decode into ASCII (Put ASCII in DL, scancode in DH)


	;; Decode table entry -- handle A_ASCII, A_FUNCTION, A_STRING(?)...
	cmp byte [esi], A_ASCII
	jnz	.2
	mov	dl, [esi+4]	;(dl: use 8-bit codes)		;"old"
	mov	dh, bl		;dh = scancode			;"old"
;	mov	edx, [esi+4]	;(dx/edx: use 16-bit codes)	;"NEW"
	jmp short .add

.2	cmp byte [esi], A_FUNCTION
	jnz	.3
	mov	eax, [esi+4]
	or	eax, eax
;	jz	.3		;Unmapped key: ignore
	pchr '<'
	phex eax
	pchr '>'
;	call	eax

.3	jmp short .end

;; Get next slot in keystroke buffer
.add	mov	eax, [keybuf_head]
	mov	ecx, eax		;ECX = keybuf_head
	add	eax, byte 2		;EAX = keybuf_head+2
	cmp	eax, keybuf.end		;Past end of buffer?
	jb	.nowrap
	mov	eax, keybuf		;Wrap around
.nowrap	cmp	eax, [keybuf_tail]		; Full? (keybuf_head+2 = keybuf_tail)
	jne short .store		; then ignore further keypresses
	pchr	7			; Give notice ("beep")
	jmp short .end
.store:	mov	[ecx], dx		; Store new character
	mov	[keybuf_head], eax		; keybuf_head = keybuf_head+2 (w/ wrap)
.end:	popa
	ret



	global add_key
add_key:
;;
;; Add a character to the keystroke buffer
;;
;; Input:
;;	character
;;
	pusha
	mov	edx, [esp+9*4]
	jmp short process_keystroke.add



	global readkey
readkey:
;
; Get a character from the typeahead buffer. (Doesn't wait)
;
; Returns a character in AX (0 if the buffer is empty)
; Returns 0 if no keys are available
;
	push	ebx

	mov	ebx, [keybuf_tail]		;EBX = keybuf_tail
	cmp	ebx, [keybuf_head]
	jne	.avail			;Keypress available? (keybuf_head != keybuf_tail)
	xor	eax, eax		;No? Return 0
	jmp short .end
.avail
	mov	eax, [ebx]		;AX = key
	add	ebx, byte 2		;EBX = keybuf_tail + 2
	cmp	ebx, keybuf.end		;Past end of buffer?
	jb	.nowrap
	mov	ebx, keybuf		;Wrap around to beginning
.nowrap
	mov	[keybuf_tail], ebx		;keybuf_tail = keybuf_tail+2
.end
	pop	ebx
	ret



	global getc
getc:
;
; Waits for a keypress and returns a character in EAX.
;
; Note: This would be the place to idle the current task and let other stuff
; run until someone hits a key :)
;
.loop
	call	readkey
	or	ax, ax	;Repeat until we get something :)
	jz	.loop
	ret



	global readline
readline:
;;
;; Reads a line of input.
;;
;; Input:
%define .addr	ebp+3*4	; Address of buffer to put string in
%define .len	ebp+2*4	; Maximum length to read in
;;
;; Output: (analogous to input)
;;	.addr			; Address (no change)
;;	.len			; Number of characters _actually read_
;;
	push	ebp
	mov	ebp, esp
	mpush	ebx, edx

	mov	ebx, [.addr]	;eax = buffer address
	xor	edx, edx	;ebx+edx = current position
.loop	call	getc
	cmp	al, 13		;CR = done
	je	.end
	cmp	al, 10		;LF = done
	je	.end
	cmp	al, 8		;Backspace
	je	.bs

	cmp	edx, [.len]	;At end of buffer?
	jae	.loop		; then ignore further keystrokes

	mov	[ebx+edx], al
	pchr	eax
	inc	edx
	jmp	.loop

.end	cr
	mov	[ebx+edx], byte 0	; null terminator
	mov	[.len], edx		; return actual length
	mpop	ebx, edx
	leave
	ret

.bs	or	edx, edx
	jz	.loop
	dec	edx
	pchr	eax
	jmp	.loop



SECTION .data align=8

keymap:	incbin "keymap"

keybuf_head:	dd	keybuf
keybuf_tail:	dd	keybuf

pause_key_loop:	db	0		; Special handling for PAUSE
extended_flag:	db	0		; 1 means "previous byte was E0"

mods	db	0	;Modifier bitmap

;; Conversion table: Extended scancodes ==> "fake" scancodes
xlat	;;;;;;;	Enter rCtrl /   rAlt Home Up  PgUp Left Right End Down PgDn Ins Del Macro
.from	db	28,   29,   53, 56,  71,  72, 73,  75,  77,   79, 80,  81,  82, 83, 111
.to	db	96,   97,   98, 100, 102,103, 104, 105, 106,  107,108, 109, 110,111,112
.end



SECTION .bss

; Typeahead buffer (a circular queue)
keybuf:
	resb	64
.end

; Table of which keys are up/down (0 = up)
key_table:
	resb	128

kt_cols	resd	1
